"""
Label smoothing
"""

from __future__ import annotations
from typing import Optional, Union, Sequence
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["label_smoothing", "smooth_one_hot", "label_smoothed_log_prob_gradient"]


def label_smoothing(prob: Tensor, smoothing: Union[Tensor, float], *, axis: Optional[Dim] = None) -> Tensor:
    """
    Label smoothing, often used for cross entropy.

    In case of sparse data, it will become dense (via :func:`smooth_one_hot`)
    and the target label will get probability (1 - smoothing).
    """
    if not axis:
        assert prob.feature_dim or prob.sparse_dim
        axis = prob.feature_dim or prob.sparse_dim
    if prob.sparse_dim:
        assert prob.sparse_dim == axis
        return rf.smooth_one_hot(prob, label_prob=1.0 - smoothing)
    else:
        assert axis in prob.dims_set
        # Make it consistent to the sparse case.
        # Value of 1.0 should result in (1 - smoothing).
        # Value of 0.0 should result in smoothing / (dim - 1).
        # Sum over all should still remain 1.0.
        dim = axis.dimension
        floor_prob = smoothing / (dim - 1)
        factor = 1.0 - dim * floor_prob
        # Case for prob[i] == 0 is clear.
        # Case for prob[i] == 1: 1 - dim * floor_prob + floor_prob = 1 + (1 - dim) * floor_prob = 1 - smoothing
        # Sum over all: 1 - dim * floor_prob + floor_prob * dim = 1
        return prob * factor + floor_prob


def smooth_one_hot(source: Tensor, *, label_prob: Union[Tensor, float]) -> Tensor:
    """
    Smooth variant of :func:`one_hot`.
    Uses ``label_prob`` for the labels and ``(1 - label_prob) / (dim - 1)`` for the remaining values.
    This is used for label smoothing.
    """
    assert source.sparse_dim
    if source.sparse_dim.dimension is None:
        raise NotImplementedError(f"smooth_one_hot({source}) not implemented for dynamic dims")
    return rf.sparse_to_dense(
        source, label_value=label_prob, other_value=(1.0 - label_prob) / (source.sparse_dim.dimension - 1)
    )


def label_smoothed_log_prob_gradient(
    log_prob: Tensor,
    smoothing: Union[Tensor, float],
    *,
    axis: Optional[Dim] = None,
    exclude_labels: Optional[Sequence[int]] = None,
) -> Tensor:
    """
    :param log_prob: shape [...,D] (not necessarily the same as loss)
    :param smoothing: smoothing factor, for :func:`label_smoothing`
    :param axis: label axis. uses feature_dim by default
    :param exclude_labels: list of labels to exclude from smoothing (e.g. blank)

    Assume some cross-entropy-like loss:

      loss = - sum_i target_prob[i] * log_prob[i] .

    The sum is over the label indices i (corresponding to the ``axis`` argument).
    Then the gradient of loss w.r.t. log_prob[i] is:

      grad_logprob[i] loss = -target_prob[i] .

    We assume that the negative gradient is a probability distribution, and apply :func:`label_smoothing` on it.
    More specifically, we apply the same scale and shift as in the :func:`label_smoothing` function
    via :func:`scaled_gradient`.

    Just as a side remark: assume

      log_prob = log_softmax(z) .

    The gradient of log_softmax is:

      grad_z[j] log_prob[i] = delta(i==j) - softmax(z)[j] .

    Then the gradient w.r.t. z[j] is:

      grad_z[j] loss = sum_i (grad_logprob[i] loss) (grad_z[j] logprob[i])
                     = sum_i -target_prob[i] delta(i==j) + target_prob[i] softmax(z)[j]
                     = -target_prob[j] + (sum_i target_prob[i]) softmax(z)[j]
                     = softmax(z)[j] - target_prob[j]    # assuming (sum_i target_prob[i]) == 1

    """
    if not axis:
        assert log_prob.feature_dim
        axis = log_prob.feature_dim
    # See formula above for label_smoothing.
    dim = axis.dimension
    floor_prob = smoothing / (dim - 1)
    factor = 1.0 - dim * floor_prob
    if exclude_labels:
        indices = rf.range_over_dim(axis)
        mask = True
        for label in exclude_labels:
            mask = mask & (indices != label)
        factor = rf.where(mask, factor, 1.0)
        floor_prob = rf.where(mask, floor_prob, 0.0)
    # The gradient is expected to be the negative target prob, thus negative floor_prob.
    # The gradient is expected to be 0. for masked frames, thus the clipping logic.
    return rf.scaled_gradient_ext(log_prob, scale=factor, shift=-floor_prob, scale_shift_by_sum_over_axis=axis)



#!/usr/bin/env python3

"""
Runs pylint on the RETURNN source code.
"""

import os
import sys
import subprocess

my_dir = os.path.dirname(os.path.abspath(__file__))
base_dir = os.path.dirname(my_dir)
sys.path.insert(0, base_dir)
os.chdir(base_dir)

from returnn.util import better_exchook  # noqa
from returnn.util.basic import pip_install, pip_check_is_installed  # noqa

py = sys.executable


def setup():
    """
    Some generic setup.
    """
    # travis_fold: https://github.com/travis-ci/travis-ci/issues/1065
    print("travis_fold:start:script.install")
    for pkg in ["pylint", "better-exchook"]:
        if not pip_check_is_installed(pkg):
            pip_install(pkg)
    print("travis_fold:end:script.install")


def main():
    """
    Main entry point.
    """
    setup()
    from lint_common import ignore_count_for_files, find_all_py_source_files

    color = better_exchook.Color()
    num_relevant_files_with_errors = 0
    for rel_filename in find_all_py_source_files():
        print("travis_fold:start:pylint.%s" % rel_filename)
        extra_args = []
        if "/" in rel_filename and not rel_filename.startswith("returnn/"):  # demos or tools
            extra_args += [r"--module-rgx=([a-z_][a-z0-9_\-]*)$"]
            extra_args += ["--disable=wrong-import-position"]  # for now
        proc = subprocess.Popen(["pylint", rel_filename] + extra_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, _ = proc.communicate()
        stdout = stdout.decode("utf8")
        file_has_errors = proc.returncode != 0
        print(
            color.color(
                "File: %s" % rel_filename,
                color="black" if (rel_filename in ignore_count_for_files or not file_has_errors) else "red",
            )
        )
        if "EXCEPTION" in stdout and "RecursionError" in stdout:
            # https://github.com/PyCQA/pylint/issues/1452
            # https://github.com/PyCQA/astroid/issues/437
            # Don't print full stdout. It will spam too much.
            print("PyLint issue #1452 triggered. https://github.com/PyCQA/pylint/issues/1452")
        elif rel_filename in ignore_count_for_files:
            print(stdout[:1000])
            print("... (ignored further output; file is not relevant)")
        else:
            print(stdout)
        print("Return code:", proc.returncode)
        if file_has_errors:
            if rel_filename in ignore_count_for_files:
                print("The inspection reports for this file are currently ignored.")
            else:
                print(color.color("The inspection reports for this file are fatal!", color="red"))
                num_relevant_files_with_errors += 1
        print("travis_fold:end:pylint.%s" % rel_filename)
    if num_relevant_files_with_errors:
        print("Num relevant files with errors:", num_relevant_files_with_errors)
        sys.exit(1)


if __name__ == "__main__":
    better_exchook.install()
    main()



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_stft():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: rf.Tensor, *, in_spatial_dim: Dim) -> Tuple[Tensor, Dim, Dim]:
            return rf.stft(x, in_spatial_dim=in_spatial_dim, frame_step=3, frame_length=5, fft_length=6)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, out_spatial_dim, out_dim = model(extern_data["data"], in_spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_mel_filterbank():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    feat_dim = Dim(10, name="mel")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, source: rf.Tensor, *, in_spatial_dim: Dim) -> Tuple[Tensor, Dim]:
            # log mel filterbank features
            source, in_spatial_dim, in_dim_ = rf.stft(
                source, in_spatial_dim=in_spatial_dim, frame_step=3, frame_length=5, fft_length=6
            )
            source = rf.abs(source) ** 2.0
            source = rf.audio.mel_filterbank(source, in_dim=in_dim_, out_dim=feat_dim, sampling_rate=16)
            return source, in_spatial_dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, out_spatial_dim = model(extern_data["data"], in_spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, feat_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)
    # Run again to specifically test the caching logic.
    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)



"""
tests for returnn.torch.frontend
"""

import _setup_test_env  # noqa

import torch
import pytest

from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


# Tensor.__eq__ is disabled as per the following error in some TF tests:
# AssertionError: unhashable type: 'Tensor'.
# See CI https://github.com/rwth-i6/returnn/actions/runs/4406240591
def test_compare_eq():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a == b
    result_alt1 = rf.compare(a, "==", b)
    result_alt2 = rf.compare(a, "eq", b)

    assert result.raw_tensor.tolist() == [False, True, False]
    assert result_alt1.raw_tensor.tolist() == [False, True, False]
    assert result_alt2.raw_tensor.tolist() == [False, True, False]


def test_compare_ne():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a != b
    result_alt1 = rf.compare(a, "!=", b)
    result_alt2 = rf.compare(a, "<>", b)
    result_alt3 = rf.compare(a, "not_equal", b)

    assert result.raw_tensor.tolist() == [True, False, True]
    assert result_alt1.raw_tensor.tolist() == [True, False, True]
    assert result_alt2.raw_tensor.tolist() == [True, False, True]
    assert result_alt3.raw_tensor.tolist() == [True, False, True]


def test_compare_lt():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a < b
    result_alt1 = rf.compare(a, "<", b)
    result_alt2 = rf.compare(a, "less", b)

    assert result.raw_tensor.tolist() == [False, False, True]
    assert result_alt1.raw_tensor.tolist() == [False, False, True]
    assert result_alt2.raw_tensor.tolist() == [False, False, True]


def test_compare_le():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a <= b
    result_alt1 = rf.compare(a, "<=", b)
    result_alt2 = rf.compare(a, "less_equal", b)

    assert result.raw_tensor.tolist() == [False, True, True]
    assert result_alt1.raw_tensor.tolist() == [False, True, True]
    assert result_alt2.raw_tensor.tolist() == [False, True, True]


def test_compare_gt():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a > b
    result_alt1 = rf.compare(a, ">", b)
    result_alt2 = rf.compare(a, "greater", b)

    assert result.raw_tensor.tolist() == [True, False, False]
    assert result_alt1.raw_tensor.tolist() == [True, False, False]
    assert result_alt2.raw_tensor.tolist() == [True, False, False]


def test_compare_ge():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a >= b
    result_alt1 = rf.compare(a, ">=", b)
    result_alt2 = rf.compare(a, "greater_equal", b)

    assert result.raw_tensor.tolist() == [True, True, False]
    assert result_alt1.raw_tensor.tolist() == [True, True, False]
    assert result_alt2.raw_tensor.tolist() == [True, True, False]


def test_combine_add_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a + b
    result_alt1 = rf.combine(a, "+", b)
    result_alt2 = rf.combine(a, "add", b)

    assert result.raw_tensor.tolist() == pytest.approx([3, 4, 5])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([3, 4, 5])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([3, 4, 5])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_add_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a + b
    result_alt1 = rf.combine(a, "+", b)
    result_alt2 = rf.combine(a, "add", b)

    assert result.raw_tensor.tolist() == pytest.approx([3.0, 4.0, 5.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([3.0, 4.0, 5.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([3.0, 4.0, 5.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_add_number_to_tensor():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor(3.0)

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[], dtype="float32")

    result = a + b
    result_alt1 = rf.combine(a, "+", b)
    result_alt2 = rf.combine(a, "add", b)

    assert result.raw_tensor.tolist() == pytest.approx([5.0, 5.0, 5.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([5.0, 5.0, 5.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([5.0, 5.0, 5.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_sub_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a - b
    result_alt1 = rf.combine(a, "-", b)
    result_alt2 = rf.combine(a, "sub", b)

    assert result.raw_tensor.tolist() == pytest.approx([1, 0, -1])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([1, 0, -1])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([1, 0, -1])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_sub_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a - b
    result_alt1 = rf.combine(a, "-", b)
    result_alt2 = rf.combine(a, "sub", b)

    assert result.raw_tensor.tolist() == pytest.approx([1.0, 0.0, -1.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([1.0, 0.0, -1.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([1.0, 0.0, -1.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_mul_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a * b
    result_alt1 = rf.combine(a, "*", b)
    result_alt2 = rf.combine(a, "mul", b)

    assert result.raw_tensor.tolist() == pytest.approx([2, 4, 6])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2, 4, 6])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2, 4, 6])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_mul_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a * b
    result_alt1 = rf.combine(a, "*", b)
    result_alt2 = rf.combine(a, "mul", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 6.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 6.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 6.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_truediv_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    # Until type promotion logic is implemented, we don't allow this.
    with pytest.raises(ValueError):
        a / b
    with pytest.raises(ValueError):
        rf.combine(a, "/", b)
    with pytest.raises(ValueError):
        rf.combine(a, "truediv", b)


def test_combine_truediv_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a / b
    result_alt1 = rf.combine(a, "/", b)
    result_alt2 = rf.combine(a, "truediv", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 2.0 / 3.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 2.0 / 3.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 2.0 / 3.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_floordiv_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a // b
    result_alt1 = rf.combine(a, "//", b)
    result_alt2 = rf.combine(a, "floordiv", b)

    assert result.raw_tensor.tolist() == pytest.approx([2, 1, 0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2, 1, 0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2, 1, 0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_floordiv_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a // b
    result_alt1 = rf.combine(a, "//", b)
    result_alt2 = rf.combine(a, "floordiv", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 0.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 0.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 0.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_mod_int_tensors():
    a_raw = torch.tensor([2, 2, 2, 17])
    b_raw = torch.tensor([1, 2, 3, 4])

    feature_dim = Dim(4)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a % b
    result_alt1 = rf.combine(a, "%", b)
    result_alt2 = rf.combine(a, "mod", b)

    assert result.raw_tensor.tolist() == pytest.approx([0, 0, 2, 1])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([0, 0, 2, 1])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([0, 0, 2, 1])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_mod_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0, 17.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0, 4.0])

    feature_dim = Dim(4)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a % b
    result_alt1 = rf.combine(a, "%", b)
    result_alt2 = rf.combine(a, "mod", b)

    assert result.raw_tensor.tolist() == pytest.approx([0.0, 0.0, 2.0, 1.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([0.0, 0.0, 2.0, 1.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([0.0, 0.0, 2.0, 1.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_pow_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a**b
    result_alt1 = rf.combine(a, "**", b)
    result_alt2 = rf.combine(a, "pow", b)

    assert result.raw_tensor.tolist() == pytest.approx([2, 4, 8])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2, 4, 8])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2, 4, 8])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_pow_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a**b
    result_alt1 = rf.combine(a, "**", b)
    result_alt2 = rf.combine(a, "pow", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 8.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 8.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 8.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_max():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = rf.combine(a, "max", b)
    result_alt1 = rf.combine(a, "maximum", b)

    assert result.raw_tensor.tolist() == [2.0, 2.0, 3.0]
    assert result_alt1.raw_tensor.tolist() == [2.0, 2.0, 3.0]


def test_combine_min():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = rf.combine(a, "min", b)
    result_alt1 = rf.combine(a, "minimum", b)

    assert result.raw_tensor.tolist() == [1.0, 2.0, 2.0]
    assert result_alt1.raw_tensor.tolist() == [1.0, 2.0, 2.0]



# Also see test_SprintDataset.py.

from __future__ import annotations

import os
import sys
import _setup_test_env  # noqa
import unittest
import numpy
import tempfile
import contextlib
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_not_in, assert_true, assert_false
from returnn.datasets.generating import Task12AXDataset, DummyDataset, DummyDatasetMultipleSequenceLength
from returnn.engine.batch import Batch
from returnn.datasets.basic import Dataset
from returnn.util.basic import NumbersDict
from returnn.util import better_exchook


def test_Task12AXDataset_deepcopy():
    from copy import deepcopy

    dataset = Task12AXDataset(num_seqs=10)
    dataset = deepcopy(dataset)
    dataset.init_seq_order(1)
    n = dataset.num_seqs
    for i in range(n):
        dataset.load_seqs(i, i + 1)
        targets = dataset.get_data(i, "classes")
        print(targets)
    assert not dataset.is_less_than_num_seqs(n)


def test_Task12AXDataset_inf():
    dataset = Task12AXDataset(num_seqs=float("inf"))
    dataset.init_seq_order(1)
    n = 10
    for i in range(n):
        dataset.load_seqs(i, i + 1)
        targets = dataset.get_data(i, "classes")
        print(targets)
    assert dataset.is_less_than_num_seqs(n)


def test_Task12AXDataset_random():
    dataset = Task12AXDataset(num_seqs=10, seq_ordering="random")
    dataset.init_seq_order(1)
    n = dataset.num_seqs
    for i in range(n):
        dataset.load_seqs(i, i + 1)
        targets = dataset.get_data(i, "classes")
        print(targets)
    assert not dataset.is_less_than_num_seqs(n)


def test_generate_batches():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=20)
    dataset.init_seq_order(1)
    batch_gen = dataset.generate_batches(recurrent_net=False, max_seqs=2, batch_size=5)
    while batch_gen.has_more():
        batch_gen.peek_next_n(1)
        batch_gen.advance(1)


def test_generate_batches_recurrent():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=20)
    dataset.init_seq_order(1)
    batch_gen = dataset.generate_batches(recurrent_net=True, max_seqs=2, batch_size=5)
    while batch_gen.has_more():
        batch_gen.peek_next_n(1)
        batch_gen.advance(1)


def test_iterate_seqs_no_chunking_1():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
    dataset.chunk_step = 0
    dataset.chunk_size = 0
    dataset.init_seq_order(1)
    seqs = list(dataset.iterate_seqs())
    assert_equal(len(seqs), 2)
    assert_equal(seqs[0], (0, 0, 11))  # seq-idx, start-frame, end-frame
    assert_equal(seqs[1], (1, 0, 11))


def test_iterate_seqs_chunking_1():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
    dataset.chunk_step = 5
    dataset.chunk_size = 10
    dataset.init_seq_order(1)
    seqs = list(dataset.iterate_seqs())
    for s in seqs:
        print(s)
    assert_equal(len(seqs), 6)
    assert_equal(seqs[0], (0, 0, 10))  # seq-idx, start-frame, end-frame
    assert_equal(seqs[1], (0, 5, 11))
    assert_equal(seqs[2], (0, 10, 11))
    assert_equal(seqs[3], (1, 0, 10))
    assert_equal(seqs[4], (1, 5, 11))
    assert_equal(seqs[5], (1, 10, 11))


def test_iterate_seqs_chunking_varying_sequence_length():
    dataset = DummyDatasetMultipleSequenceLength(
        input_dim=2, output_dim=3, num_seqs=2, seq_len={"data": 24, "classes": 12}
    )
    dataset.chunk_size = {"data": 12, "classes": 6}
    dataset.chunk_step = {"data": 6, "classes": 3}
    dataset.init_seq_order(1)
    seqs = list(dataset.iterate_seqs())
    for s in seqs:
        print(s)
    assert_equal(len(seqs), 8)
    assert_equal(seqs[0], (0, NumbersDict({"data": 0, "classes": 0}), NumbersDict({"data": 12, "classes": 6})))
    assert_equal(seqs[1], (0, NumbersDict({"data": 6, "classes": 3}), NumbersDict({"data": 18, "classes": 9})))
    assert_equal(seqs[2], (0, NumbersDict({"data": 12, "classes": 6}), NumbersDict({"data": 24, "classes": 12})))
    assert_equal(seqs[3], (0, NumbersDict({"data": 18, "classes": 9}), NumbersDict({"data": 24, "classes": 12})))
    assert_equal(seqs[4], (1, NumbersDict({"data": 0, "classes": 0}), NumbersDict({"data": 12, "classes": 6})))
    assert_equal(seqs[5], (1, NumbersDict({"data": 6, "classes": 3}), NumbersDict({"data": 18, "classes": 9})))
    assert_equal(seqs[6], (1, NumbersDict({"data": 12, "classes": 6}), NumbersDict({"data": 24, "classes": 12})))
    assert_equal(seqs[7], (1, NumbersDict({"data": 18, "classes": 9}), NumbersDict({"data": 24, "classes": 12})))


def test_iterate_seqs_custom_chunking():
    default_key = "data"
    chunk_step = 5
    chunk_size = 10

    def _custom_chunking_func(dataset, seq_idx_start, **_kwargs):
        assert isinstance(dataset, Dataset)
        seq_idx = seq_idx_start
        while dataset.is_less_than_num_seqs(seq_idx):
            length = dataset.get_seq_length(seq_idx)
            t = NumbersDict.constant_like(0, numbers_dict=length)
            while length[default_key] > t[default_key]:
                chunk_start = NumbersDict(t)
                chunk_end = NumbersDict.min([t + chunk_size, length])
                yield seq_idx, chunk_start, chunk_end
                t += chunk_step
            seq_idx += 1

    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
    dataset.custom_chunking_func = _custom_chunking_func
    dataset.init_seq_order(1)
    seqs = list(dataset.iterate_seqs())
    for s in seqs:
        print(s)
    assert_equal(len(seqs), 6)
    assert_equal(seqs[0], (0, 0, 10))  # seq-idx, start-frame, end-frame
    assert_equal(seqs[1], (0, 5, 11))
    assert_equal(seqs[2], (0, 10, 11))
    assert_equal(seqs[3], (1, 0, 10))
    assert_equal(seqs[4], (1, 5, 11))
    assert_equal(seqs[5], (1, 10, 11))


def test_batches_recurrent_1():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
    dataset.init_seq_order(1)
    dataset.chunk_size = 10
    dataset.chunk_step = 5
    batch_gen = dataset.generate_batches(recurrent_net=True, max_seqs=1, batch_size=20)
    all_batches = []
    " :type: list[Batch] "
    while batch_gen.has_more():
        (batch,) = batch_gen.peek_next_n(1)
        assert_is_instance(batch, Batch)
        print("batch:", batch)
        print("batch seqs:", batch.seqs)
        all_batches.append(batch)
        batch_gen.advance(1)

    # Each batch will have 1 batch-slice (max_seqs) and up to 10 frames (chunk_size).
    # For each seq, we get 3 chunks (chunk_step 5 for 11 frames).
    # Thus, 6 batches.
    assert_equal(len(all_batches), 6)

    assert_equal(all_batches[0].start_seq, 0)
    assert_equal(all_batches[0].end_seq, 1)  # exclusive
    assert_equal(len(all_batches[0].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[0].seqs[0].seq_idx, 0)
    assert_equal(all_batches[0].seqs[0].seq_start_frame, 0)
    assert_equal(all_batches[0].seqs[0].seq_end_frame, 10)
    assert_equal(all_batches[0].seqs[0].frame_length, 10)
    assert_equal(all_batches[0].seqs[0].batch_slice, 0)
    assert_equal(all_batches[0].seqs[0].batch_frame_offset, 0)

    assert_equal(all_batches[1].start_seq, 0)
    assert_equal(all_batches[1].end_seq, 1)  # exclusive
    assert_equal(len(all_batches[1].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[1].seqs[0].seq_idx, 0)
    assert_equal(all_batches[1].seqs[0].seq_start_frame, 5)
    assert_equal(all_batches[1].seqs[0].seq_end_frame, 11)
    assert_equal(all_batches[1].seqs[0].frame_length, 6)
    assert_equal(all_batches[1].seqs[0].batch_slice, 0)
    assert_equal(all_batches[1].seqs[0].batch_frame_offset, 0)

    assert_equal(all_batches[2].start_seq, 0)
    assert_equal(all_batches[2].end_seq, 1)  # exclusive
    assert_equal(len(all_batches[2].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[2].seqs[0].seq_idx, 0)
    assert_equal(all_batches[2].seqs[0].seq_start_frame, 10)
    assert_equal(all_batches[2].seqs[0].seq_end_frame, 11)
    assert_equal(all_batches[2].seqs[0].frame_length, 1)
    assert_equal(all_batches[2].seqs[0].batch_slice, 0)
    assert_equal(all_batches[2].seqs[0].batch_frame_offset, 0)

    assert_equal(all_batches[3].start_seq, 1)
    assert_equal(all_batches[3].end_seq, 2)  # exclusive
    assert_equal(len(all_batches[3].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[3].seqs[0].seq_idx, 1)
    assert_equal(all_batches[3].seqs[0].seq_start_frame, 0)
    assert_equal(all_batches[3].seqs[0].seq_end_frame, 10)
    assert_equal(all_batches[3].seqs[0].frame_length, 10)
    assert_equal(all_batches[3].seqs[0].batch_slice, 0)
    assert_equal(all_batches[3].seqs[0].batch_frame_offset, 0)

    # ...


def test_batches_non_recurrent_1():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=2, seq_len=11)
    dataset.init_seq_order(1)
    batch_gen = dataset.generate_batches(recurrent_net=False, max_seqs=2, batch_size=5)
    all_batches = []  # type: list[Batch]
    while batch_gen.has_more():
        (batch,) = batch_gen.peek_next_n(1)
        assert_is_instance(batch, Batch)
        print("batch:", batch)
        print("batch seqs:", batch.seqs)
        all_batches.append(batch)
        batch_gen.advance(1)

    # Each batch will have 5 frames (batch_size), not more, i.e. a single seq.
    # There are 2 * 11 frames in total, so 5 batches, because we concat the 2 seqs, in the non-recurrent case.
    assert_equal(len(all_batches), 5)

    assert_equal(all_batches[0].start_seq, 0)
    assert_equal(all_batches[0].end_seq, 1)  # exclusive
    assert_equal(len(all_batches[0].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[0].seqs[0].seq_idx, 0)
    assert_equal(all_batches[0].seqs[0].seq_start_frame, 0)
    assert_equal(all_batches[0].seqs[0].seq_end_frame, 5)
    assert_equal(all_batches[0].seqs[0].frame_length, 5)
    assert_equal(all_batches[0].seqs[0].batch_slice, 0)
    assert_equal(all_batches[0].seqs[0].batch_frame_offset, 0)

    assert_equal(all_batches[1].start_seq, 0)
    assert_equal(all_batches[1].end_seq, 1)  # exclusive
    assert_equal(len(all_batches[1].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[1].seqs[0].seq_idx, 0)
    assert_equal(all_batches[1].seqs[0].seq_start_frame, 5)
    assert_equal(all_batches[1].seqs[0].seq_end_frame, 10)
    assert_equal(all_batches[1].seqs[0].frame_length, 5)
    assert_equal(all_batches[1].seqs[0].batch_slice, 0)
    assert_equal(all_batches[1].seqs[0].batch_frame_offset, 0)

    assert_equal(all_batches[2].start_seq, 0)
    assert_equal(all_batches[2].end_seq, 2)  # exclusive. now both seq 0 and 1
    assert_equal(len(all_batches[2].seqs), 2)  # two copies, BatchSeqCopyPart
    assert_equal(all_batches[2].seqs[0].seq_idx, 0)
    assert_equal(all_batches[2].seqs[0].seq_start_frame, 10)
    assert_equal(all_batches[2].seqs[0].seq_end_frame, 11)
    assert_equal(all_batches[2].seqs[0].frame_length, 1)
    assert_equal(all_batches[2].seqs[0].batch_slice, 0)
    assert_equal(all_batches[2].seqs[0].batch_frame_offset, 0)
    assert_equal(all_batches[2].seqs[1].seq_idx, 1)
    assert_equal(all_batches[2].seqs[1].seq_start_frame, 0)
    assert_equal(all_batches[2].seqs[1].seq_end_frame, 4)
    assert_equal(all_batches[2].seqs[1].frame_length, 4)
    assert_equal(all_batches[2].seqs[1].batch_slice, 0)
    assert_equal(all_batches[2].seqs[1].batch_frame_offset, 1)

    assert_equal(all_batches[3].start_seq, 1)
    assert_equal(all_batches[3].end_seq, 2)  # exclusive
    assert_equal(len(all_batches[3].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[3].seqs[0].seq_idx, 1)
    assert_equal(all_batches[3].seqs[0].seq_start_frame, 4)
    assert_equal(all_batches[3].seqs[0].seq_end_frame, 9)
    assert_equal(all_batches[3].seqs[0].frame_length, 5)
    assert_equal(all_batches[3].seqs[0].batch_slice, 0)
    assert_equal(all_batches[3].seqs[0].batch_frame_offset, 0)

    assert_equal(all_batches[4].start_seq, 1)
    assert_equal(all_batches[4].end_seq, 2)  # exclusive
    assert_equal(len(all_batches[4].seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(all_batches[4].seqs[0].seq_idx, 1)
    assert_equal(all_batches[4].seqs[0].seq_start_frame, 9)
    assert_equal(all_batches[4].seqs[0].seq_end_frame, 11)
    assert_equal(all_batches[4].seqs[0].frame_length, 2)
    assert_equal(all_batches[4].seqs[0].batch_slice, 0)
    assert_equal(all_batches[4].seqs[0].batch_frame_offset, 0)


def test_batches_context_window():
    context_window = 2
    ctx_lr = context_window - 1
    ctx_left = ctx_lr // 2
    ctx_right = ctx_lr - ctx_left

    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=1, seq_len=11, context_window=context_window)
    dataset.init_seq_order(1)
    dataset.chunk_size = 5
    dataset.chunk_step = 5
    batch_gen = dataset.generate_batches(recurrent_net=True, max_seqs=1, batch_size=20)
    all_batches = []  # type: list[Batch]
    while batch_gen.has_more():
        (batch,) = batch_gen.peek_next_n(1)
        assert_is_instance(batch, Batch)
        print("batch:", batch)
        print("batch seqs:", batch.seqs)
        all_batches.append(batch)
        batch_gen.advance(1)

    # Each batch will have 1 batch-slice (max_seqs) and up to 10 frames (chunk_size).
    # For each seq, we get 3 chunks (chunk_step 5 for 11 frames).
    # Thus, 3 batches.
    assert_equal(len(all_batches), 3)
    b0, b1, b2 = all_batches
    assert isinstance(b0, Batch)
    assert isinstance(b1, Batch)
    assert isinstance(b2, Batch)

    assert_equal(b0.start_seq, 0)
    assert_equal(b0.end_seq, 1)  # exclusive
    assert_equal(len(b0.seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(b0.seqs[0].seq_idx, 0)
    assert_equal(b0.seqs[0].seq_start_frame["classes"], 0)
    assert_equal(b0.seqs[0].seq_end_frame["classes"], 5)
    assert_equal(b0.seqs[0].frame_length["classes"], 5)
    assert_equal(b0.seqs[0].seq_start_frame["data"], 0 - ctx_left)
    assert_equal(b0.seqs[0].seq_end_frame["data"], 5 + ctx_right)
    assert_equal(b0.seqs[0].frame_length["data"], 5 + ctx_lr)
    assert_equal(b0.seqs[0].batch_slice, 0)
    assert_equal(b0.seqs[0].batch_frame_offset, 0)

    assert_equal(b1.start_seq, 0)
    assert_equal(b1.end_seq, 1)  # exclusive
    assert_equal(len(b1.seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(b1.seqs[0].seq_idx, 0)
    assert_equal(b1.seqs[0].seq_start_frame["classes"], 5)
    assert_equal(b1.seqs[0].seq_end_frame["classes"], 10)
    assert_equal(b1.seqs[0].frame_length["classes"], 5)
    assert_equal(b1.seqs[0].seq_start_frame["data"], 5 - ctx_left)
    assert_equal(b1.seqs[0].seq_end_frame["data"], 10 + ctx_right)
    assert_equal(b1.seqs[0].frame_length["data"], 5 + ctx_lr)
    assert_equal(b1.seqs[0].batch_slice, 0)
    assert_equal(b1.seqs[0].batch_frame_offset, 0)

    assert_equal(b2.start_seq, 0)
    assert_equal(b2.end_seq, 1)  # exclusive
    assert_equal(len(b2.seqs), 1)  # 1 BatchSeqCopyPart
    assert_equal(b2.seqs[0].seq_idx, 0)
    assert_equal(b2.seqs[0].seq_start_frame["classes"], 10)
    assert_equal(b2.seqs[0].seq_end_frame["classes"], 11)
    assert_equal(b2.seqs[0].frame_length["classes"], 1)
    assert_equal(b2.seqs[0].seq_start_frame["data"], 10 - ctx_left)
    assert_equal(b2.seqs[0].seq_end_frame["data"], 11 + ctx_right)
    assert_equal(b2.seqs[0].frame_length["data"], 1 + ctx_lr)
    assert_equal(b2.seqs[0].batch_slice, 0)
    assert_equal(b2.seqs[0].batch_frame_offset, 0)


def test_task12ax_window():
    from returnn.datasets.generating import Task12AXDataset

    window = 3
    dataset_kwargs = dict(num_seqs=10)
    dataset1 = Task12AXDataset(**dataset_kwargs)
    dataset2 = Task12AXDataset(window=window, **dataset_kwargs)
    input_dim = dataset1.num_inputs
    dataset1.initialize()
    dataset2.initialize()
    dataset1.init_seq_order(epoch=1)
    dataset2.init_seq_order(epoch=1)
    dataset1.load_seqs(0, 1)
    dataset2.load_seqs(0, 1)
    assert_equal(dataset1.get_data_dim("data"), input_dim)
    assert_equal(dataset2.get_data_dim("data"), input_dim * window)
    data1 = dataset1.get_data(0, "data")
    data2 = dataset2.get_data(0, "data")
    seq_len = data1.shape[0]
    assert_equal(data1.shape, (seq_len, input_dim))
    assert_equal(data2.shape, (seq_len, window * input_dim))
    data2a = data2.reshape(seq_len, window, input_dim)
    print("data1:")
    print(data1)
    print("data2:")
    print(data2)
    print("data1[0]:")
    print(data1[0])
    print("data2[0]:")
    print(data2[0])
    print("data2a[0,0]:")
    print(data2a[0, 0])
    assert_equal(list(data2a[0, 0]), [0] * input_dim)  # zero-padded left
    assert_equal(list(data2a[0, 1]), list(data1[0]))
    assert_equal(list(data2a[0, 2]), list(data1[1]))
    assert_equal(list(data2a[1, 0]), list(data1[0]))
    assert_equal(list(data2a[1, 1]), list(data1[1]))
    assert_equal(list(data2a[1, 2]), list(data1[2]))
    assert_equal(list(data2a[-1, 2]), [0] * input_dim)  # zero-padded right


def test_get_seq_order():
    dataset = Dataset()
    num_seqs = 30

    def get_seq_len(i):
        return i**2 % 17  # some dummy lengths

    for seq_ordering in [
        "default",
        "default_every_n:5",
        "sorted",
        "sorted_reverse",
        "random:3",
        "laplace:3",
        "laplace:.10",
        "sort_bin_shuffle:3",
        "sort_bin_shuffle_x2:.10",
    ]:

        dataset.seq_ordering = seq_ordering

        # test full epoch
        dataset.partition_epoch = 1
        epoch = 3
        seq_index = dataset.get_seq_order_for_epoch(epoch, num_seqs, get_seq_len)

        assert isinstance(seq_index, (list, range, numpy.ndarray))
        assert len(set(seq_index)) == num_seqs  # right number of sequences, no duplicates

        # test partitioned epoch
        partition_epoch = 4
        dataset.partition_epoch = partition_epoch
        all_partitions_seq_index = []
        for epoch in range(1, partition_epoch + 1):
            partition_seq_index = dataset.get_seq_order_for_epoch(epoch, num_seqs, get_seq_len)
            all_partitions_seq_index += list(partition_seq_index)

        # Make sure partitions combined result in full epoch. This tests the random seed of Dataset which should be
        # fixed across partitions.
        assert set(all_partitions_seq_index) == set(seq_index)


@contextlib.contextmanager
def create_ogg_zip_txt_only_dataset(*, text: str = "hello world", seq_tag: str = "sequence0.wav"):
    import zipfile
    from returnn.datasets.audio import OggZipDataset

    with tempfile.NamedTemporaryFile(suffix=".zip") as tmp_zip_file, tempfile.NamedTemporaryFile(
        suffix=".txt"
    ) as tmp_vocab_file:
        with zipfile.ZipFile(tmp_zip_file.name, "w") as zip_file:
            zip_file.writestr(
                os.path.basename(tmp_zip_file.name)[:-4] + ".txt",
                repr([{"text": text, "duration": 2.3, "file": seq_tag}]),
            )
        vocab = {"@": 2, " ": 1, ".": 0}
        vocab.update({chr(i): i - ord("a") + 3 for i in range(ord("a"), ord("z") + 1)})
        tmp_vocab_file.write(repr(vocab).encode("utf8"))
        tmp_vocab_file.flush()

        dataset = OggZipDataset(
            path=tmp_zip_file.name,
            audio=None,
            targets={"class": "CharacterTargets", "vocab_file": tmp_vocab_file.name, "seq_postfix": [0]},
        )
        dataset.initialize()
        yield dataset


def test_OggZipDataset():
    from returnn.datasets.audio import OggZipDataset

    _demo_txt = "some utterance text"

    with create_ogg_zip_txt_only_dataset(text=_demo_txt) as dataset:
        assert isinstance(dataset, OggZipDataset)
        dataset.init_seq_order(epoch=1)
        dataset.load_seqs(0, 1)
        raw = dataset.get_data(0, "raw")
        orth = dataset.get_data(0, "orth")
        classes = dataset.get_data(0, "classes")
        print("raw:", raw)
        print("orth:", orth)
        print("classes:", classes)
        assert isinstance(raw, numpy.ndarray) and raw.dtype.name.startswith("str") and raw.shape == ()
        raw_ = raw.item()
        assert isinstance(raw_, str) and raw_ == _demo_txt
        assert isinstance(orth, numpy.ndarray) and orth.dtype == numpy.uint8 and orth.ndim == 1
        orth_ = orth.tostring()
        assert orth_.decode("utf8") == _demo_txt
        assert isinstance(classes, numpy.ndarray) and classes.dtype == numpy.int32 and classes.ndim == 1
        classes_ = "".join([dataset.targets.id_to_label(c) for c in classes])
        assert classes_ == _demo_txt + "."


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
Utilities for dimension tags, dimensions, axes.
"""


from __future__ import annotations
from typing import Optional, Union, TypeVar, Sequence, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import get_backend_by_tensor, global_backend

T = TypeVar("T")

__all__ = ["range_over_dim", "replace_dim", "dim_match_priority_when_needed", "num_elements_of_shape"]


def range_over_dim(dim: Dim, *, dtype: Optional[str] = None) -> Tensor[T]:
    """
    :param dim:
    :param dtype:
    :return: tensor with shape [dim]
    """
    if dim.dyn_size_ext:
        backend = get_backend_by_tensor(dim.dyn_size_ext, fallback=global_backend)
    else:
        backend = global_backend
    return backend.range_over_dim(dim, dtype=dtype)


def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Optional[Dim] = None) -> Tuple[Tensor, Dim]:
    """
    :param source:
    :param in_dim:
    :param out_dim:
    :return: source with in_dim replaced by out_dim, and new out_dim.
        this does not work for the sparse_dim. see :func:`set_sparse_dim` for that case.
    """
    if not out_dim:
        out_dim = in_dim.copy(same_as_self=False, description="new-dim")
    # noinspection PyProtectedMember
    return source._raw_backend.replace_dim(source, in_dim=in_dim, out_dim=out_dim), out_dim


def dim_match_priority_when_needed(dim: Dim, *other_dims: Dim) -> Dim:
    """
    :return: maybe copy of dim with higher match_priority if needed to distinguish from other_dims

    Why or when is this needed?

    For activation values, this should never be needed,
    and all dims should be unique.

    In case of self-attention, the standard way is to create a separate distinct dim
    to perform the attention reduction over.
    See :class:`SelfAttention`.

    However, in case of weight matrices, it is not unusual to have the same dim for both the input and output,
    so a square weight matrix.
    When reduction is performed in :func:`matmul`, we want to match the input feature dim
    to the dim in the weight matrix with higher match priority.

    So :func:`dim_match_priority_when_needed` would be applied on the input feature dim.

    https://github.com/rwth-i6/returnn/pull/871
    https://github.com/rwth-i6/returnn_common/issues/17#issuecomment-997463222
    """
    if dim in other_dims:
        return dim.copy(match_priority=1)
    return dim


def num_elements_of_shape(dims: Sequence[Dim]) -> Union[int, Tensor]:
    """
    :param dims:
    :return: num elements of a tensor of shape dims, properly considering masking
    """
    if all(dim.is_static() for dim in dims):
        n = 1
        for dim in dims:
            n *= dim.dimension
        return n

    n = 1
    dims = list(dims)
    dims.sort(key=lambda dim: -dim.dyn_size_ext.batch_ndim if dim.dyn_size_ext else 0)
    while dims:
        dim = dims.pop(0)
        if dim.is_static():
            n *= dim.dimension
            continue
        # E.g. dyn_size_ext is shape [B], and self has shape [B,T].
        # Due to the sorting of dims above, dims will be [T,B], and we will first process T.
        # We want to sum over dyn_size_ext, but then we need to remove the other dims it covers.
        for dim_ in dim.dyn_size_ext.dims:
            assert dim_ in dims  # num elements not really well-defined then
            assert not dim_.need_masking()  # not implemented
            dims.remove(dim_)
        n_ = rf.reduce_sum(dim.dyn_size_ext, axis=dim.dyn_size_ext.dims)
        n *= n_
    return n
